"use strict";

exports.__esModule = true;
exports.usePress = usePress;

var _react = _interopRequireDefault(require("react"));

var _checkTargetContainsSelection = _interopRequireDefault(require("../../modules/checkTargetContainsSelection"));

var _supportsDOM = _interopRequireDefault(require("../../modules/supportsDOM"));

var _modality = require("../../modules/modality");

var _addEvent = require("../../modules/addEvent");

var _useEvent = require("../use-event");

var _useKeyboard = require("../use-keyboard");

var _useLayoutEffect = _interopRequireDefault(require("../../modules/useLayoutEffect"));

var _useMergeRefs = require("../use-merge-refs");

var _useResponder = require("../use-responder");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * 
 */
var emptyObject = {};
var MOVE_THRESHOLD_PX = 10;
var DEFAULT_LONG_PRESS_DELAY_MS = 500;
var captureOptions = {
  capture: true
};

function getTouchFromResponderEvent(event) {
  var _event$nativeEvent = event.nativeEvent,
      changedTouches = _event$nativeEvent.changedTouches,
      touches = _event$nativeEvent.touches;

  if (touches != null && touches.length > 0) {
    return touches[0];
  }

  if (changedTouches != null && changedTouches.length > 0) {
    return changedTouches[0];
  }

  return event.nativeEvent;
}

function isKeyActivator(key) {
  return key === 'Enter' || isKeySpacebar(key);
}

function isKeySpacebar(key) {
  return key === ' ' || key === 'Spacebar';
}

function isTargetDisabled(target) {
  return target.disabled || target.getAttribute('aria-disabled') === 'true';
}

function isTargetNativelyInteractive(target) {
  var href = target.href,
      tagName = target.tagName;
  return tagName === 'A' && href != null || tagName === 'BUTTON' || tagName === 'INPUT' || tagName === 'SELECT' || tagName === 'TEXTAREA';
}

function shouldSpacebarActivate(event) {
  var currentTarget = event.currentTarget;
  var role = currentTarget.getAttribute('role');
  var spacebarActivates = role === 'button' || role === 'checkbox' || role === 'menuitem' || role === 'menuitemcheckbox' || role === 'menuitemradio' || role === 'radio' || role === 'switch';
  return spacebarActivates;
}

function useClick(listener) {
  var addClickListener = (0, _useEvent.useEvent)('click');
  return _react.default.useCallback(function (target) {
    if (target != null) {
      addClickListener(target, listener);
    }
  }, [addClickListener, listener]);
}

function useContextMenu(listener) {
  var addContextMenuListener = (0, _useEvent.useEvent)('contextmenu');
  return _react.default.useCallback(function (target) {
    if (target != null) {
      addContextMenuListener(target, listener);
    }
  }, [addContextMenuListener, listener]);
}

function usePress(config) {
  var _ref = config || emptyObject,
      delayLongPress = _ref.delayLongPress,
      disabled = _ref.disabled,
      onLongPress = _ref.onLongPress,
      onPress = _ref.onPress,
      onPressCancel = _ref.onPressCancel,
      onPressChange = _ref.onPressChange,
      onPressEnd = _ref.onPressEnd,
      onPressStart = _ref.onPressStart,
      touchAction = _ref.touchAction;

  var longPressDelay = delayLongPress || DEFAULT_LONG_PRESS_DELAY_MS;

  var stateRef = _react.default.useRef({
    active: false,
    initialX: null,
    initialY: null,
    isLongPress: false,
    longPressTimeout: null
  });

  var targetRef = _react.default.useRef(null);

  var clickRef = useClick(function (e) {
    var state = stateRef.current;
    var target = e.currentTarget;

    if (isTargetDisabled(target)) {
      e.stopPropagation();
    } else if (!disabled) {
      e.stopPropagation();
      var deltaX = state.initialX != null ? state.initialX - e.pageX : 0;
      var deltaY = state.initialY != null ? state.initialY - e.pageY : 0; // The browser will select text beneath a pointer when it performs a rapid double-tap.
      // We only want to cancel press if the user moves the pointer to select text.

      var hasSelection = (0, _checkTargetContainsSelection.default)(target);
      var didSelectText = Math.hypot(deltaX, deltaY) > MOVE_THRESHOLD_PX && hasSelection; // Click triggers onPress if there is no active longpress, and no text
      // selection occured during the press. Otherwise, it is cancelled.

      if (state.isLongPress || didSelectText) {
        e.preventDefault();
      } else if (onPress != null) {
        onPress(e);
      }
    }
  });
  var contextMenuRef = useContextMenu(function (e) {
    var target = e.currentTarget;

    if (isTargetDisabled(target) || !disabled && onLongPress != null && (0, _modality.getActiveModality)() === 'touch') {
      // Context menu is cancelled for touch pointers if onLongPress is defined.
      // Context menu is cancelled if the element or hook is disabled
      e.stopPropagation();
      e.preventDefault();
    }
  });
  /**
   * Pointers
   */

  var responderRef = (0, _useResponder.useResponder)({
    onStartShouldSetResponder: function onStartShouldSetResponder(e) {
      var target = e.currentTarget;
      return !disabled && !isTargetDisabled(target);
    },
    onResponderGrant: function onResponderGrant(e) {
      var state = stateRef.current;
      var touch = getTouchFromResponderEvent(e);
      state.active = true;
      state.isLongPress = false;
      state.initialX = touch.pageX;
      state.initialY = touch.pageY;

      if (onPressStart != null) {
        onPressStart(e);
      }

      if (onPressChange != null) {
        onPressChange(true);
      }

      if (onLongPress != null) {
        state.longPressTimeout = setTimeout(function () {
          if (state.active === true) {
            state.isLongPress = true;
            onLongPress();
          }
        }, longPressDelay);
      }
    },
    onResponderMove: function onResponderMove(e) {
      var state = stateRef.current;
      var touch = getTouchFromResponderEvent(e);
      var deltaX = state.initialX != null ? state.initialX - touch.pageX : 0;
      var deltaY = state.initialY != null ? state.initialY - touch.pageY : 0; // If the pointer moves more than the threshold, cancel longpress

      if (onLongPress != null && Math.hypot(deltaX, deltaY) > MOVE_THRESHOLD_PX) {
        clearTimeout(state.longPressTimeout);
        state.longPressTimeout = null;
      }
    },
    onResponderRelease: function onResponderRelease(e) {
      var state = stateRef.current;

      if (onPressChange != null) {
        onPressChange(false);
      }

      if (onPressEnd != null) {
        onPressEnd(e);
      }

      clearTimeout(state.longPressTimeout);
      state.active = false;
      state.longPressTimeout = null;
    },
    onResponderTerminate: function onResponderTerminate(e) {
      var state = stateRef.current;

      if (onPressChange != null) {
        onPressChange(false);
      }

      if (onPressCancel != null) {
        onPressCancel(e);
      }

      clearTimeout(state.longPressTimeout);
      state.active = false;
      state.isLongPress = false;
      state.longPressTimeout = null;
    },
    onResponderTerminationRequest: function onResponderTerminationRequest(e) {
      if (onLongPress != null && e.nativeEvent.type === 'contextmenu' && (0, _modality.getActiveModality)() === 'touch') {
        return false;
      }

      return true;
    },
    touchAction: touchAction
  });
  /**
   * Keyboard
   */

  var keyboardRef = (0, _useKeyboard.useKeyboard)({
    disabled: disabled,
    onKeyDown: function onKeyDown(e) {
      var key = e.key;
      var state = stateRef.current;

      if (isTargetDisabled(e.currentTarget) && !isKeySpacebar(key)) {
        e.stopPropagation();
      } else if (isKeyActivator(key)) {
        // Stop the spacebar potentially scrolling the window
        if (isKeySpacebar(key)) {
          if (shouldSpacebarActivate(e)) {
            e.preventDefault();
          } else {
            return;
          }
        }

        e.stopPropagation();

        if (state.active !== true) {
          if (onPressStart != null) {
            onPressStart(e);
          }

          if (onPressChange != null) {
            onPressChange(true);
          }

          state.active = true;
        }
      }
    }
  }); // The 'keyup' listener is attached to the document to account for situations where
  // focus is moved during 'keydown', resulting in 'keyup' not occuring on the target.
  // In that case, we want to clear the pressed state on 'keyup' without triggering 'onPress'.
  // Use the capture phase on the off-chance that the 'keyup' target prevents propagation.

  (0, _useLayoutEffect.default)(function () {
    if ((0, _supportsDOM.default)()) {
      var target = targetRef.current;
      var state = stateRef.current;

      if (target != null) {
        var listener = function listener(e) {
          var key = e.key;

          if (target != null && state.active && isKeyActivator(key)) {
            if (onPressChange != null) {
              onPressChange(false);
            }

            if (onPressEnd != null) {
              onPressEnd(e);
            }

            if (e.target === target) {
              // Only stop propagation if the key is being handled by the original target
              e.stopPropagation();

              if (onPress != null && !isTargetNativelyInteractive(e.target)) {
                onPress(e);
              }
            }

            state.active = false;
          }
        };

        var removeKeyUpCaptureListener = (0, _addEvent.addEvent)(document, 'keyup', listener, captureOptions);
        return removeKeyUpCaptureListener;
      }
    }
  }, [onPressChange, onPressEnd, onPress, stateRef, targetRef]);
  return (0, _useMergeRefs.useMergeRefs)(responderRef, clickRef, contextMenuRef, keyboardRef, targetRef);
}