/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @flow
 */

import React from 'react';
import { addEvent, removeEvent } from '#internal/addEvent';

type FocusConfig = {
  disabled?: ?boolean,
  onBlur?: ?(e: any) => void,
  onBlurCapture?: ?(e: any) => void,
  onFocus?: ?(e: any) => void,
  onFocusCapture?: ?(e: any) => void,
  onFocusChange?: ?(isFocused: boolean, e: any) => void
};

/**
 * useFocusEvents implementation
 */

const bubbleOptions = { passive: true };
const captureOptions = { capture: true, passive: true };

export function useFocus(config: FocusConfig) {
  const {
    disabled,
    onBlur,
    onBlurCapture,
    onFocus,
    onFocusCapture,
    onFocusChange
  } = config;

  const attachedNode = React.useRef(null);

  return React.useCallback(
    (target: any) => {
      function blurListener(e) {
        if (disabled) {
          return;
        }
        if (onBlur != null) {
          onBlur(e);
        }
        if (onFocusChange != null) {
          onFocusChange(false, e);
        }
      }

      function focusListener(e) {
        if (disabled) {
          return;
        }
        if (onFocus != null) {
          onFocus(e);
        }
        if (onFocusChange != null) {
          onFocusChange(true, e);
        }
      }

      if (target !== null) {
        attachedNode.current = target;
        addEvent(target, 'blur', blurListener, bubbleOptions);
        addEvent(target, 'focus', focusListener, bubbleOptions);
        if (onBlurCapture != null) {
          addEvent(
            target,
            'blur',
            (!disabled && onBlurCapture) || null,
            captureOptions
          );
        }
        if (onFocusCapture != null) {
          addEvent(
            target,
            'focus',
            (!disabled && onFocusCapture) || null,
            captureOptions
          );
        }
      } else if (attachedNode.current != null) {
        const node = attachedNode.current;
        removeEvent(node, 'blur', blurListener, bubbleOptions);
        removeEvent(node, 'focus', focusListener, bubbleOptions);
        removeEvent(node, 'blur', onBlurCapture, captureOptions);
        removeEvent(node, 'focus', onFocusCapture, captureOptions);
        attachedNode.current = null;
      }
    },
    [disabled, onBlur, onBlurCapture, onFocus, onFocusCapture, onFocusChange]
  );
}
