/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @flow
 */

import React from 'react';
import { addEvent, removeEvent } from '#internal/addEvent';

/**
 * Types
 */

type KeyboardEventsConfig = {
  disabled?: ?boolean,
  onKeyDown?: (e: any) => void,
  onKeyDownCapture?: (e: any) => void,
  onKeyUp?: (e: any) => void,
  onKeyUpCapture?: (e: any) => void
};

/**
 * useKeyboardEvents implementation
 */

const captureOptions = { capture: true };

// TODO: https://github.com/w3c/uievents/issues/202
export function useKeyboard(config: KeyboardEventsConfig) {
  const {
    disabled,
    onKeyDown,
    onKeyDownCapture,
    onKeyUp,
    onKeyUpCapture
  } = config;

  const attachedNode = React.useRef(null);

  return React.useCallback(
    (target: any) => {
      const createKeyListener = (listener) => {
        return function keyListener(e) {
          if (disabled) {
            return;
          }
          // Ignore composition events
          if (!e.isComposing && e.keyCode !== 229) {
            if (listener != null) {
              listener(e);
            }
          }
        };
      };

      const keyDownListener = createKeyListener(onKeyDown);
      const keyUpListener = createKeyListener(onKeyUp);
      const keyDownCaptureListener = createKeyListener(onKeyDownCapture);
      const keyUpCaptureListener = createKeyListener(onKeyUpCapture);

      if (target != null) {
        addEvent(target, 'keydown', keyDownListener);
        addEvent(target, 'keyup', keyUpListener);
        addEvent(target, 'keydown', keyDownCaptureListener, captureOptions);
        addEvent(target, 'keyup', keyUpCaptureListener, captureOptions);
      } else if (attachedNode.current != null) {
        const node = attachedNode.current;
        removeEvent(node, 'keydown', keyDownListener);
        removeEvent(node, 'keyup', keyUpListener);
        removeEvent(node, 'keydown', keyDownCaptureListener, captureOptions);
        removeEvent(node, 'keyup', keyUpCaptureListener, captureOptions);
        attachedNode.current = null;
      }
    },
    [disabled, onKeyDown, onKeyDownCapture, onKeyUp, onKeyUpCapture]
  );
}
